--- **LibBattleTag-1.0** provides conversion between BattleTags and player names
-- TODO: Add Description
--
-- @usage
-- local LBN = LibStub("LibBattleTag-1.0")
-- LBN:Enable("some unique id")
-- local playerName = LBN:GetPlayerNameFromBattlTag("Areko#0000")
-- local battleTag = LBN:GetBattleTagFromPlayerName("Areko")
-- local battleTag = LBN:GetBattleTagFromPlayerName("Areko-Alleria")
-- local playerInfo = LBN:GetPlayerInfoFromBattleTag("Areko#0000")
-- local playerInfo = LBN:GetPlayerInfoFromPlayerName("Areko")
-- local playerInfo = LBN:GetPlayerInfoFromPlayerName("Areko-Alleria")
--
-- @usage
-- local addon = LibStub("LibBattleTag-1.0"):Embed({})
-- addon:LBN_Enable()
-- local playerName = addon:GetPlayerNameFromBattlTag("Areko#0000")
-- local battleTag = addon:GetBattleTagFromPlayerName("Areko")
-- local battleTag = addon:GetBattleTagFromPlayerName("Areko-Alleria")
-- local playerInfo = addon:GetPlayerInfoFromBattleTag("Areko#0000")
-- local playerInfo = addon:GetPlayerInfoFromPlayerName("Areko")
-- local playerInfo = addon:GetPlayerInfoFromPlayerName("Areko-Alleria")
--
-- @usage
-- local ADDONNAME, addon = ...
-- LibStub("AceAddon-3.0"):NewAddon(addon, ADDONNAME, "LibBattleTag-1.0")
-- function addon:DoThings()
--     local playerName = self:GetPlayerNameFromBattlTag("Areko#0000")
--     local battleTag = self:GetBattleTagFromPlayerName("Areko")
--     local battleTag = self:GetBattleTagFromPlayerName("Areko-Alleria")
--     local playerInfo = self:GetPlayerInfoFromBattleTag("Areko#0000")
--     local playerInfo = self:GetPlayerInfoFromPlayerName("Areko")
--     local playerInfo = self:GetPlayerInfoFromPlayerName("Areko-Alleria")
-- end
--
-- @class file
-- @name LibBattleTag-1.0

-- luacheck: globals LibStub
assert(LibStub, "LibBattleTag-1.0 requires LibStub.")

local lib, old = LibStub:NewLibrary("LibBattleTag-1.0", 1)
if not lib then return end

-- Lua API
local type, pairs, assert, next =
      type, pairs, assert, next
-- local getmetatable, setmetatable = getmetatable, setmetatable
local string_find, string_sub, string_gsub =
      string.find, string.sub, string.gsub
local string_lower = string.utf8_lower or string.lower
local string_upper = string.utf8_upper or string.upper

-- WoW API
local BNGetNumFriends, BNGetFriendInfo, BNGetGameAccountInfo, BNGetFriendInfoByID
    = BNGetNumFriends, BNGetFriendInfo, BNGetGameAccountInfo, BNGetFriendInfoByID

-- WoW API constants
local BNET_CLIENT_WOW = BNET_CLIENT_WOW

-- WoW GlobalStrings
local UNKNOWN = UNKNOWN

-- Constants
local HOMEREALM = string_gsub(GetRealmName(), "%s", "")
local HOMEREALM_LOWERED = string_lower(HOMEREALM)
local NAME_REALM_DELIMITER = "-"
local NAME_REALM_DELIMITER_FILTER = "%s*"..NAME_REALM_DELIMITER.."%s*"

-- Local functions
local firstCase, uniformName, duplicate
local updateConvertTables, updatePlayerInfo, updateOnlineBattleTagFriends, removePlayerInfo

-- Internal lib data
lib.frame = lib.frame or CreateFrame("Frame")
lib.embeds = lib.embeds or {}
lib.enablers = lib.enablers or {}
lib.internalData = lib.internalData or {}
lib.valueToPlayerInfo = lib.valueToPlayerInfo or {}
lib.knownRealmNames = lib.knownRealmNames or {}
lib.knownRealmNames[HOMEREALM_LOWERED] = HOMEREALM

lib.frame:SetScript("OnEvent", function(self, event, ...)
    if lib[event] then
        lib[event](lib, event, ...)
    end
end)


-----------------------------------------------------------
-- Utile functions

do-- firstCase
    -- Pattern to match the first character for UTF8 strings
    local MULTIBYTE_FIRST_CHAR = "^([\192-\255]?%a?[\128-\191]*)"
    function firstCase(s)
        return string_gsub(s, MULTIBYTE_FIRST_CHAR, string_upper)
    end
end
--[=[
local function TitleCase(s)

    return string_gsub(firstCase(s), "[^%s+]..-", firstCase)
end
--]=]

function uniformName(name, realm, overrideRealm)
    name = name or UNKNOWN
    realm = realm or HOMEREALM

    local startPos, endPos = string_find(name, NAME_REALM_DELIMITER_FILTER)
    if startPos then
        realm = overrideRealm and realm or string_sub(name, endPos+1)
        realm = string_gsub(realm, "%s", "")
        name = string_sub(name, 0, startPos-1)
    end

    return name..NAME_REALM_DELIMITER..realm, name, realm
end

function duplicate(value)
    --local valueType = type(value)
    local duplicatedValue

    --if valueType == "table" then
    if type(value) == "table" then
        -- replace, when we need metatable support
        -- duplicatedValue = setmetatable({}, getmetatable(value))
        duplicatedValue = {}
        for k,v in pairs(value) do
            -- NOTE: duplicated k?
            duplicatedValue[k] = duplicate(v)
        end
    else
        duplicatedValue = value
    end

    return duplicatedValue
end


-----------------------------------------------------------
-- Lib methods

--- TODO: Add Title/Description
-- @param target (any object)   - Used to identified different enabler
function lib:Enable(target)
    if not next(lib.enablers) then
        updateOnlineBattleTagFriends()
        lib.frame:RegisterEvent("BN_FRIEND_INFO_CHANGED")
    end
    if target == nil then
        target = self
    end
    lib.enablers[target] = true
end

--- TODO: Add Title/Description
-- @param target (any object)   - Used to identified different disabler
function lib:Disable(target)
    if target == nil and self ~= lib then
        target = self
    end

    if target then
        lib.enablers[target] = nil

        if not next(lib.enablers) then
            lib.frame:UnregisterEvent("BN_FRIEND_INFO_CHANGED")
        end
    end
end

--- TODO: Add Title/Description
-- @name PlayerInfo
-- @type table
-- @entry name (String)      - Player name e.g: Areko
-- @entry realm (String)     - e.g: Alleria
-- @entry fullName (String)  - Fullname with Realm e.g: Areko-Alleria
-- @entry battleTag (String) - BattleTag e.g: Areko#0000

--- TODO: Add Tile/Description
-- @param BattleTag (String)    - Case sensitive or lower case
-- @return name (String or nil) - Returns nil when unknown BattleTag
function lib:GetPlayerNameFromBattlTag(tag)
    assert(type(tag) == "string", "TODO: error message")
    assert(string_find(tag, '#'), "TODO: error message")

    local player = lib.valueToPlayerInfo[tag]
    local name = player and (player.fullName or player.name or UNKNOWN) or nil

    return name
end

--- TODO: Add Tile/Description
-- @param BattleTag (String)            - Case sensitive or lower case
-- @return PlayerInfo (table or nil)    - Table with information about the player @see PlayerInfo
--                                        or nil when unknown BattleTag
function lib:GetPlayerInfoFromBattleTag(tag)
    assert(type(tag) == "string", "TODO: error message")
    assert(string_find(tag, '#'), "TODO: error message")
    -- `x or nil` is here to guarantee nil as result when the playerInfo is false
    local playerInfo = lib.valueToPlayerInfo[tag] or nil

    return duplicate(playerInfo)
end

--- TODO: Add Tile/Description
-- @param Player name (String)          - Case insensitive
-- @return BattleTag (String or nil)    - BattleTag
--                                        or nil when the player is not on your BattleNet Friend list
function lib.GetBattleTagFromPlayerName(name)
    assert(type(name) == "string", "TODO: error message")
    name = string_lower( (uniformName(name)) )
    local player = lib.valueToPlayerInfo[name]

    -- `x or nil` is here to guarantee nil as result when the playerInfo is false
    local battleTag = player and player.battleTag or nil

    return battleTag
end

--- TODO: Add Tile/Description
-- @param Player name (string)          - Case insensitive
-- @return PlayerInfo (table or nil)    - Table with information about the player @see PlayerInfo
--                                        or nil when unknown BattleTag
function lib:GetPlayerInfoFromPlayerName(name)
    assert(type(name) == "string", "TODO: error message")
    name = string_lower( (uniformName(name)) )

    -- `x or nil` is here to guarantee nil as result when the playerInfo is false
    local playerInfo = lib.valueToPlayerInfo[name] or nil

    return duplicate(playerInfo)
end


-----------------------------------------------------------
-- Internal functions

function updateOnlineBattleTagFriends()
    local numBNetTotal, numBNetOnline = BNGetNumFriends()
    for index=1, numBNetOnline do
        lib:BN_FRIEND_INFO_CHANGED("BN_FRIEND_INFO_CHANGED", index)
    end
end

-- BN_FRIEND_ACCOUNT_ONLINE -> arg1 --> BNGetFriendInfoByID(arg1)
-- BN_FRIEND_ACCOUNT_OFFLINE -> arg1 --> BNGetFriendInfoByID(arg1)
-- BN_FRIEND_INFO_CHANGED -> index --> BNGetFriendInfo(index)
-- local bnetIDAccount, accountName, battleTag, isBattleTag, characterName, bnetIDGameAccount,
--  client, isOnline, lastOnline, isBNAFK, isBNDND, broadcastText, noteText, isRIDFriend
--  = BNGetFriendInfoByID(bnetIDAccount)
-- local bnetIDAccount, accountName, battleTag, isBattleTag, characterName, bnetIDGameAccount,
--  client, isOnline, lastOnline, isBNAFK, isBNDND, broadcastText, noteText, isRIDFriend,
--  messageTime, canSoR, isReferAFriend, canSummonFriend = BNGetFriendInfo(index)
-- local hasFocus, characterName, client, realmName, realmID, faction, race, class, _, zoneName,
--  level, gameText = BNGetGameAccountInfo(bnetIDGameAccount)

do-- lib:BN_FRIEND_INFO_CHANGED()
    -- TODO: verify the way how the BN* API returns realmName
    --       and slim down the knownRealmNames table accordingly.
    local function addToKnownRealmNames(realmName)
        local realmName_api = string_gsub(realmName, "%s*", "")

        lib.knownRealmNames[realmName] = realmName_api
        lib.knownRealmNames[realmName_api] = realmName_api
        lib.knownRealmNames[string_lower(realmName)] = realmName_api
        lib.knownRealmNames[string_lower(realmName_api)] = realmName_api
    end

    -- TODO: I thought to remember that BN_FRIEND_INFO_CHANGED fires quite often with the same value,
    --       without real change. So maybe best to add a from of throttling
    -- TODO: clean up unused variables
    function lib:BN_FRIEND_INFO_CHANGED(event, friendIndex)
        if friendIndex then
            local bnetIDAccount, accountName, battleTag, isBattleTag, characterName, bnetIDGameAccount,
             client, isOnline, lastOnline, isBNAFK, isBNDND, broadcastText, noteText, isRIDFriend,
             messageTime, canSoR, isReferAFriend, canSummonFriend = BNGetFriendInfo(friendIndex)

            if client == BNET_CLIENT_WOW and bnetIDGameAccount then
                local hasFocus, realmName, realmID, faction, race, class, _, zoneName, level, gameText

                hasFocus, characterName, client, realmName, realmID, faction, race, class, _,
                 zoneName, level, gameText = BNGetGameAccountInfo(bnetIDGameAccount)

                if not lib.knownRealmNames[realmName] then
                    addToKnownRealmNames(realmName)
                end

                local _, name = uniformName(characterName, realmName)
                updatePlayerInfo(battleTag, name, realmName)
            end
        end
    end
end

function lib:BN_FRIEND_ACCOUNT_OFFLINE(event, bnetIDAccount)
    if bnetIDAccount then
        local accountName, battleTag
        bnetIDAccount, accountName, battleTag = BNGetFriendInfoByID(bnetIDAccount)

        local playerInfo = lib.internalData[battleTag or ""]
        removePlayerInfo(playerInfo)
    end
end

function updateConvertTables(specificPlayerInfo)
    if specificPlayerInfo then
        local battleTag         = specificPlayerInfo.battleTag
        local battleTag_lowered = string_lower(battleTag)
        local name              = specificPlayerInfo.fullName
        local name_lowered      = string_lower(name)

        lib.valueToPlayerInfo[battleTag]            = specificPlayerInfo
        lib.valueToPlayerInfo[battleTag_lowered]    = specificPlayerInfo
        lib.valueToPlayerInfo[name]                 = specificPlayerInfo
        lib.valueToPlayerInfo[name_lowered]         = specificPlayerInfo
    else
        for battleTag_lowered, playerInfo in pairs(lib.internalData) do
            if playerInfo then
                local battleTag     = playerInfo.battleTag
                local name          = playerInfo.fullName
                local name_lowered  = string_lower(name)

                lib.valueToPlayerInfo[battleTag]            = playerInfo
                lib.valueToPlayerInfo[battleTag_lowered]    = playerInfo
                lib.valueToPlayerInfo[name]                 = playerInfo
                lib.valueToPlayerInfo[name_lowered]         = playerInfo
            end
        end
    end
end

function updatePlayerInfo(battleTag, name, realm)
    --[===[@debug@
    assert(battleTag, "No BattleTag.")
    assert(string_find(battleTag, '#'), "Incorrect BattleTag.")
    assert(name and realm, "No name nor realm.")
    --@end-debug@]===]
    if battleTag and string_find(battleTag, "#") then
        local battleTag_lowered = string_lower(battleTag)
        local playerInfo = lib.internalData[battleTag_lowered]
        if not playerInfo then
            if not name or not realm then
                return
            end

            playerInfo = {}
        elseif not name and not realm then
            return
        end

        local fullName
        if realm then
            realm = lib.knownRealmNames[string_lower(realm)] or string_gsub(realm, "%s*", "")
        else
            realm = playerInfo.realm
        end
        name = name and firstCase(name) or playerInfo.name
        fullName, name, realm = uniformName(name, realm, true)


        playerInfo.battleTag = battleTag
        playerInfo.name = name
        playerInfo.realm = realm
        playerInfo.fullName = fullName

        -- Update/Save data
        lib.internalData[battleTag_lowered] = playerInfo

        updateConvertTables(playerInfo)
    end
end

do-- removePlayerInfo
    local TABLE_NAME_LIST = {
        valueToPlayerInfo = true,
        internalData = true,
    }

    -- function removePlayerInfo(playerInfo)
    --     for name in pairs(TABLE_NAME_LIST) do
    --         local table = lib[name]
    --         for id, object in pairs(table) do
    --             if object == playerInfo then
    --                 table[id] = false
    --             end
    --         end
    --     end
    -- end

    function removePlayerInfo(playerInfo)
        if playerInfo then
            for name in pairs(TABLE_NAME_LIST) do
                local table = lib[name]

                if table then
                    local key, object = next(table)

                    while key do
                        if object == playerInfo then
                            local lastKey = key
                            key, object = next(table, key)
                            table[lastKey] = nil
                        else
                            key, object = next(table, key)
                        end
                    end
                end
            end
        end
    end
end


-----------------------------------------------------------
-- Embedding

local mixins = {
    LBN_Enable = "Enable",
    LBN_Disable = "Disable",
    GetPlayerNameFromBattlTag = false,
    GetPlayerInfoFromBattleTag = false,
    GetBattleTagFromPlayerName = false,
    GetPlayerInfoFromPlayerName = false,
}
lib.OnEmbedEnable = lib.Enable
lib.OnEmbedDisable = lib.Disable
function lib:Embed(target)
    lib.embeds[target] = true
    for k, v in pairs(mixins) do
        target[k] = lib[v or k]
    end
    return target
end
for addon in pairs(lib.embeds) do
    lib:Embed(addon)
end
